"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;

var _qs = _interopRequireDefault(require("qs"));

var _extend = _interopRequireDefault(require("extend"));

var _pathToRegexp = _interopRequireDefault(require("path-to-regexp"));

var _koaBody = _interopRequireDefault(require("koa-body"));

var _utils = require("./utils");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

class Layer {
  constructor(path, methods, middlewares, opts) {
    this.opts = opts || {};
    this.name = this.opts.name || null;
    this.methods = [];
    this.pathKeys = [];
    this.parameters = [];
    this.operations = methods;
    this.stack = Array.isArray(middlewares) ? middlewares : [middlewares];
    methods.forEach(method => {
      const m = method.toUpperCase();
      this.methods.push(m); // https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods/HEAD

      if (m === 'GET') {
        this.methods.unshift('HEAD');
      }
    }); // ensure middleware is a function

    this.stack.forEach(fn => {
      const msg = `${methods} ${this.opts.name || path}: middleware must be a function`;
      (0, _utils.assert)(typeof fn === 'function', msg);
    });
    this.path = path;
    this.setPrefix(this.opts.prefix || '');
    this.setBodyParser();
  }

  getPathItem() {
    const pathTemplate = this.pathTemplate,
          parameters = this.parameters,
          operations = this.operations,
          opts = this.opts;

    if (!pathTemplate || opts.ignoreParams) {
      return {};
    }

    const path = pathTemplate();
    let tmp = {};

    if (operations.length) {
      const obj = (0, _utils.takeInOptions)(opts, 'method');

      if (parameters.length) {
        (0, _extend.default)(true, obj, {
          parameters
        });
      }

      if (this.bodySchema) {
        const map = {
          json: 'application/json',
          form: 'application/x-www-form-urlencoded',
          multipart: 'multipart/form-data'
        };
        const bodyType = opts.bodyType || ['json', 'form'];
        const requestBody = {
          content: {}
        };
        [].concat(bodyType).forEach(t => {
          const type = map[t] || t;
          requestBody.content[type] = {
            schema: this.bodySchema
          };
        });
        (0, _extend.default)(true, obj, {
          requestBody
        });
      }

      operations.forEach(m => {
        tmp[m.toLowerCase()] = obj;
      });
    } else {
      tmp = (0, _utils.takeInOptions)(opts, 'path');

      if (parameters.length) {
        (0, _extend.default)(true, tmp, {
          parameters
        });
      }
    }

    return {
      [path]: tmp
    };
  }

  setParameters() {
    const pathKeys = this.pathKeys,
          opts = this.opts;
    const params = opts.params;
    const inPath = {};
    const parameters = pathKeys.map((key, index) => {
      const name = String(key.name);
      inPath[name] = {
        index
      };
      return {
        name,
        in: 'path',
        required: !key.optional,
        schema: {
          type: 'string'
        }
      };
    });
    params && Object.keys(params).forEach(name => {
      const obj = params[name] || {};
      const param = (0, _utils.takeInOptions)(obj, 'param');
      const schema = (0, _utils.takeInOptions)(obj, 'schema');
      param.name = name;

      if (obj.type) {
        (0, _extend.default)(true, schema, (0, _utils.transformType)(obj.type));
      }

      if (Object.keys(schema).length) {
        param.schema = (0, _extend.default)(true, param.schema, schema);
      }

      const find = inPath[name];

      if (find) {
        const index = find.index;
        parameters[index] = (0, _extend.default)(true, parameters[index], param);
        (0, _utils.assert)(parameters[index].in === 'path', `${name} must be in path`);
      } else {
        parameters.push(param);
      }
    });
    this.parameters = parameters;
  }

  getParamsValidate() {
    if (this.paramsValidate) {
      return this.paramsValidate;
    }

    if (this.parameters.length === 0 || this.opts.ignoreParams) {
      return null;
    }

    const properties = {};
    const required = [];
    let hasProps = false;
    this.parameters.forEach(p => {
      if (p.schema) {
        hasProps = true;
        properties[p.name] = p.schema;

        if (p.required) {
          required.push(p.name);
        }
      }
    });
    if (!hasProps) return null;
    const validator = this.opts.validator;
    if (!validator) return null;
    this.paramsValidate = validator.compile({
      type: 'object',
      properties,
      required
    });
    return this.paramsValidate;
  }

  setBodyParser() {
    const _this$opts = this.opts,
          body = _this$opts.body,
          bodyparser = _this$opts.bodyparser;

    if (bodyparser || body) {
      this.setBodyValidate();
      const options = bodyparser === true ? {} : bodyparser;
      this.stack.unshift((0, _koaBody.default)(options));
    }
  }

  setBodyValidate() {
    const _this$opts2 = this.opts,
          body = _this$opts2.body,
          validator = _this$opts2.validator;
    if (!body || !validator) return;
    this.bodySchema = (0, _utils.propsToSchema)(body);

    if (this.bodySchema) {
      this.stack.unshift((ctx, next) => {
        if (!this.bodyValidate) {
          this.bodyValidate = validator.compile(this.bodySchema);
        }

        const throwBodyError = this.opts.throwBodyError;
        (0, _extend.default)(ctx.request.body, ctx.request.files);
        const valid = this.bodyValidate(ctx.request.body);
        ctx.bodyErrors = this.bodyValidate.errors;

        if (!valid && throwBodyError !== false) {
          if (typeof throwBodyError === 'function') {
            throwBodyError(ctx.bodyErrors);
          } else {
            (0, _utils.validateError)(ctx.bodyErrors);
          }
        }

        return next();
      });
    }
  }

  setPrefix(prefix) {
    if (this.path) {
      this.path = prefix + this.path;

      if (/.+\/$/.test(this.path)) {
        this.path = this.path.replace(/\/$/, '');
      }

      (0, _utils.debug)('defined route %s %s', this.methods, this.path);

      const tokens = _pathToRegexp.default.parse(this.path, this.opts);

      this.pathKeys = [];
      this.regexp = _pathToRegexp.default.tokensToRegExp(tokens, this.pathKeys, this.opts);

      this.url = (params, options = {}) => {
        let replace = {};

        if (Array.isArray(params)) {
          this.pathKeys.forEach((key, i) => {
            replace[key.name] = params[i];
          });
        } else {
          replace = params;
        }

        const toPath = _pathToRegexp.default.tokensToFunction(tokens);

        const base = toPath(replace, options);
        return (0, _utils.toURI)(base, options.query);
      };

      this.pathTemplate = () => {
        const obj = this.pathKeys.reduce((memo, k) => {
          memo[k.name] = k.name;
          return memo;
        }, {});
        return this.url(obj, {
          encode: v => `{${v}}`,
          noValidate: true
        });
      };

      this.setParameters();
    }

    return this;
  }

  match(path) {
    return this.regexp.test(path);
  }

  params(path, ctx) {
    // "query", "header", "path" or "cookie"
    const params = ctx.params || {};

    if (this.opts.ignoreParams) {
      return params;
    }

    const parameters = this.parameters,
          opts = this.opts;

    const paramsIn = s => parameters.filter(p => p.in === s); // handle params in header


    paramsIn('header').forEach(p => {
      params[p.name] = ctx.get(p.name);
    }); // handle params in cookie

    paramsIn('cookie').forEach(p => {
      params[p.name] = ctx.cookies.get(p.name, opts.cookieOptions);
    }); // handle params in query

    const queryParams = paramsIn('query');

    if (queryParams.length) {
      const query = (() => {
        const str = ctx.querystring;
        const c = ctx._querycache = ctx._querycache || {};
        return c[str] || (c[str] = _qs.default.parse(str, opts.queryOptions));
      })();

      queryParams.forEach(p => {
        params[p.name] = query[p.name];
      });
    } // handle params in path


    if (!ctx._pathParsed) {
      const regexp = this.regexp,
            pathKeys = this.pathKeys;
      const captures = path.match(regexp).slice(1);

      for (let len = captures.length, i = 0; i < len; i++) {
        if (pathKeys[i]) {
          const c = captures[i];
          params[pathKeys[i].name] = c ? (0, _utils.safeDecodeURIComponent)(c) : c;
        }
      }

      ctx._pathParsed = true;
    }

    const validate = this.getParamsValidate();

    if (validate) {
      const valid = validate(params);
      ctx.paramsErrors = validate.errors;

      if (!valid && opts.throwParamsError !== false) {
        if (typeof opts.throwParamsError === 'function') {
          opts.throwParamsError(ctx.paramsErrors);
        } else {
          (0, _utils.validateError)(ctx.paramsErrors);
        }
      }
    }

    return params;
  }

  param(param, fn) {
    const stack = this.stack,
          parameters = this.parameters;

    const paramFn = (ctx, next) => fn(ctx.params[param], ctx, next);

    paramFn.param = param;
    const names = parameters.map(p => p.name);
    const x = names.indexOf(param);

    if (x > -1) {
      stack.some((mw, i) => {
        if (!mw.param || names.indexOf(mw.param) > x) {
          stack.splice(i, 0, paramFn);
          return true; // break the loop
        }
      });
    }

    return this;
  }

}

exports.default = Layer;