"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.safeDecodeURIComponent = safeDecodeURIComponent;
exports.assert = assert;
exports.validateError = validateError;
exports.toURI = toURI;
exports.takeInOptions = takeInOptions;
exports.ref = ref;
exports.getMixType = getMixType;
exports.transformExtends = transformExtends;
exports.transformType = transformType;
exports.propsToSchema = propsToSchema;
exports.FORMATS = exports.TYPES = exports.debug = void 0;

var _qs = _interopRequireDefault(require("qs"));

var _extend = _interopRequireDefault(require("extend"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _objectWithoutProperties(source, excluded) { if (source == null) return {}; var target = _objectWithoutPropertiesLoose(source, excluded); var key, i; if (Object.getOwnPropertySymbols) { var sourceSymbolKeys = Object.getOwnPropertySymbols(source); for (i = 0; i < sourceSymbolKeys.length; i++) { key = sourceSymbolKeys[i]; if (excluded.indexOf(key) >= 0) continue; if (!Object.prototype.propertyIsEnumerable.call(source, key)) continue; target[key] = source[key]; } } return target; }

function _objectWithoutPropertiesLoose(source, excluded) { if (source == null) return {}; var target = {}; var sourceKeys = Object.keys(source); var key, i; for (i = 0; i < sourceKeys.length; i++) { key = sourceKeys[i]; if (excluded.indexOf(key) >= 0) continue; target[key] = source[key]; } return target; }

const debug = require('debug')('koa-mapper');

exports.debug = debug;

function safeDecodeURIComponent(text) {
  try {
    return decodeURIComponent(text);
  } catch (e) {
    return text;
  }
}

function assert(value, message) {
  if (value) return;
  throw new Error(message);
}

function validateError(errors) {
  const message = errors.map(e => {
    if (e.dataPath) {
      const key = e.dataPath.replace(/^./, '');
      return `[${key}] ${e.message}`;
    } else {
      return e.message;
    }
  }).join('\n');
  const err = new Error(message);
  err.status = 400;
  err.expose = true;
  throw err;
}

function toURI(base, query) {
  if (!query) return base;

  if (typeof query === 'string') {
    query = _qs.default.parse(query);
  }

  const str = _qs.default.stringify(query);

  if (!str) return base;

  if (base.indexOf('?') >= 0) {
    return `${base}&${str}`;
  } else {
    return `${base}?${str}`;
  }
}

function takeInOptions(opts, key) {
  const map = {
    'path': ['summary', 'description'],
    'method': ['tags', 'summary', 'description', 'externalDocs', 'responses', 'callbacks', 'deprecated', 'security', 'servers', 'requestBody'],
    'schema': ['items', 'title', 'multipleOf', 'maximum', 'exclusiveMaximum', 'minimum', 'exclusiveMinimum', 'maxLength', 'minLength', 'pattern', 'maxItems', 'minItems', 'uniqueItems', 'maxProperties', 'minProperties', 'enum', 'default', 'format'],
    'param': ['name', 'in', 'description', 'required', 'deprecated', 'allowEmptyValue', 'style', 'explode', 'allowReserved', 'schema', 'example', 'examples']
  };
  const obj = {};
  map[key].forEach(k => {
    if (opts[k] !== undefined) {
      obj[k] = opts[k];
    }
  });
  return obj;
}

const TYPES = ['array', 'boolean', 'integer', 'null', 'number', 'object', 'string'];
exports.TYPES = TYPES;
const FORMATS = ['date', 'time', 'date-time', 'regex'];
exports.FORMATS = FORMATS;

function ref(name) {
  return `#/components/schemas/${name}`;
}

function getMixType(type) {
  const types = {
    file: {
      type: 'object',
      file: true
    }
  };
  TYPES.forEach(t => {
    types[t] = {
      type: t
    };
  });
  FORMATS.forEach(t => {
    types[t] = {
      type: 'string',
      format: t,
      convert: true
    };
  });
  types['datetime'] = types['date-time'];
  const hasAnd = /&/.test(type);
  const hasOr = /\|/.test(type);

  if (hasAnd && hasOr) {
    throw new Error('& and | can only have one');
  }

  const arr = [];
  type.split(/[&|]/).forEach(t => {
    t = t.trim();

    if (t) {
      const simple = types[t.toLowerCase()];
      arr.push(simple || {
        $ref: ref(t)
      });
    }
  });

  if (arr.length > 1) {
    return hasAnd ? {
      allOf: arr
    } : {
      oneOf: arr
    };
  } else {
    return arr[0];
  }
}

function transformExtends(name) {
  const arr = name.split(/:/);
  const clz = {
    name: arr[0].trim(),
    parents: []
  };

  if (arr[1]) {
    arr[1].split(/,/).forEach(str => {
      const parent = str.trim();

      if (parent) {
        clz.parents.push({
          $ref: ref(parent)
        });
      }
    });
  }

  return clz;
}

function transformType(type) {
  type = type.trim();
  if (!type) return {};
  const arrayRE = /array\s?<([^<>]+)>/i;
  const m = arrayRE.exec(type);

  if (m) {
    const str = m[1].trim();

    if (!str) {
      return {
        type: 'array'
      };
    }

    return {
      type: 'array',
      items: getMixType(str)
    };
  } else {
    return getMixType(type);
  }
}

function propsToSchema(props, options = {}) {
  if (props && typeof props === 'string') {
    return {
      $ref: ref(props)
    };
  }

  if (props && Object.keys(props).length) {
    const properties = {};
    const requiredArr = options.required || [];
    Object.keys(props).forEach(k => {
      const _props$k = props[k],
            type = _props$k.type,
            required = _props$k.required,
            others = _objectWithoutProperties(_props$k, ["type", "required"]);

      const typeObj = transformType(type);
      properties[k] = (0, _extend.default)(true, others, typeObj);

      if (required) {
        requiredArr.push(k);
      }
    });
    const required = [...new Set(requiredArr)];
    const schema = {
      type: 'object',
      properties
    };

    if (required.length) {
      schema.required = required;
    }

    return schema;
  }

  return null;
}